/*
 * Mailfront Plugin Name: plugin-check-helo
 *
 * Reference:
 * http://www.linuxmagic.com/best_practices/valid_helo_domain.html
 *
 * This plugin is used to check helo hostname.
 * If $RELAYCLIENT or $HELO_CHECK_SKIP is set, this plugin will be skipped.
 * If sender is authenticated, this plugin will be skipped also.
 * If $HELO_CHECK_REJECT is set, sender will be rejected with reply code 554.
 * If $HELO_CHECK_DEFER is set, sender will be deferred with reply code 421.
 * NOTE: $HELO_CHECK_REJECT take preference of $HELO_CHECK_DEFER.
 * If $HELO_CHECK_PTR is set, this plugin will make sure the PTR record
 * resolves to the same given helo name.  If this is not set, this plugin will
 * only check for PTR record of the given helo name's IP exists.
 * If $HELO_CHECK_REMOTEHOST is set, this plugin will check helo hostname and
 * remotehost whether both are the same otherwise add headers or defer or
 * reject accordingly.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <resolv.h>

#include <msg/msg.h>
#include "mailfront.h"

#define eprintf(...) fprintf(stderr, __VA_ARGS__)

static response resp;

static const char *mainprog = "mailfront";
static const char *thisprog = "plugin-check-helo";
static unsigned long thispid;
static const char *local_host;
static const char *local_ip;
static const char *remote_host;
static const char *remote_ip;
static const char *relayclient;
static const char *rejectbadhelo;
static const char *deferbadhelo;
static const char *skiphelocheck;
static const char *checkremotehost;
static int outofmemory = 0;
static str errormsg;
static str headermsg;
static str respmsg;

struct HostDNSIP {
	char *hostname;
	int size;
	int ptrcount;
	int ptrcheck;
	int ipmemused;
	int ptrmemused;
	char *ip[10];
	char *ptr[10];
};

int free_HostDNSIP(struct HostDNSIP *d)
{
	int x = 0;
	int y = 0;
	if (d->size > 0) {
		if (d->ipmemused > 0) {
			y = d->ipmemused;
			if (d->ipmemused > d->size) {
			    y = d->size;
			}
			for (x = 0; x < y; x++) {
				free(d->ip[x]);
			}
		}
		if (d->ptrmemused > 0) {
			y = d->ptrmemused;
			if (d->ptrmemused > d->size) {
			    y = d->size;
			}
			for (x = 0; x < y; x++) {
				free(d->ptr[x]);
			}
		}
	}
	return 1;
}

char *ip_ptr_lookup(char *ip)
{
	struct hostent *hent;
	struct in_addr addr;

	if(!inet_aton(ip, &addr))
		return ip;

	char *hostname = ip;

	if(strchr(ip, ':')) {
		if((hent = gethostbyaddr((char *)&(addr.s_addr),
					 sizeof(addr.s_addr),
					 AF_INET6))) {
			strcpy(hostname, hent->h_name);
		}
	} else {
		if((hent = gethostbyaddr((char *)&(addr.s_addr),
					 sizeof(addr.s_addr),
					 AF_INET))) {
			strcpy(hostname, hent->h_name);
		}
	}
	ip = hostname;
	return ip;
}

int is_valid_ipv4(const char *s)
{
	unsigned int n1, n2, n3, n4;

	if (sscanf(s, "%3u.%3u.%3u.%3u", &n1, &n2, &n3, &n4) != 4) return 0;

	if ((n1 != 0) && (n1 <= 255) && (n2 <= 255) && (n3 <= 255)
	    && (n4 <= 255)) {
		char buf[64];
		sprintf(buf, "%u.%u.%u.%u", n1, n2, n3, n4);
		if(strcmp(buf, s)) return 0;
		return 1;
	}
	return 0;
}

void eprintf_logname()
{
	if (!thispid) {
		thispid = (unsigned long)getpid();
	}
	eprintf("%s-%s[%lu]: ", mainprog, thisprog, thispid);
}

void str_tolower(const char *a, char *b) {
	int i, len = strlen(a);
	for(i = 0; i < len; i++) {
		b[i] = (char)tolower(a[i]);
	}
	b[len] = '\0';
}

/*
 * return 1 if hostname DNS lookup has A record IPv4 or IPv6
 * otherwise return 0 if failed or out of memory which will increment
 * static int outofmemory
 */
static int check_hostname(const char *hostname, struct HostDNSIP *s)
{
	struct addrinfo hints, *res, *result;
	int ret = 0;
	int found = 0;
	int i = 0;
	if (hostname) {
		if (s->hostname) {
			eprintf_logname();
			eprintf("(%s:%d) s->hostname already set to %s\n",
				__FILE__, __LINE__, s->hostname);
			return 1;
		}
		memset(&hints, 0, sizeof(hints));
		hints.ai_family = AF_UNSPEC; // Handle IPv4 and IPv6
		hints.ai_socktype = SOCK_STREAM;
		ret = getaddrinfo(hostname, NULL, &hints, &result);
		if(ret) {
			eprintf_logname();
			eprintf("(%s:%d) hostname=%s errno=%d errmsg=%s\n",
				__FILE__, __LINE__, hostname, ret,
				gai_strerror(ret));
			char *t;
			sprintf(t, "hostname=%s errno=%d errmsg=%s", hostname,
				   ret, gai_strerror(ret));
			str_init(&errormsg);
			str_cats(&errormsg, t);
			return 0;
		} else {
			for (res = result; res != NULL; res = res->ai_next) {
				char ipstr[64];
				int x = 0;
				int saved = 1;
				struct sockaddr_in *ipv =
				    (struct sockaddr_in *)res->ai_addr;
				void * addr = &(ipv->sin_addr);
				inet_ntop(res->ai_family, addr, ipstr,
					  sizeof(ipstr));
				if (strcmp(ipstr, hostname)) {
					s->hostname = (char *)hostname;
					found = 1;
				}
				if (i > 0) {
					for (x = 0; x < i; x++) {
						if (strcmp(s->ip[x],
						    ipstr) == 0) {
							saved = 0;
						}
					}
				}
				if (saved > 0) {
					s->ip[i] = strdup(ipstr);
					if (s->ip[i] == NULL) {
						outofmemory++;
						return 0;
					}
					s->ipmemused += 1;
					s->ptr[i] = strdup(
						    ip_ptr_lookup(ipstr));
					if (s->ptr[i] == NULL) {
						outofmemory++;
						return 0;
					}
					s->ptrmemused += 1;
					i++;
				}
			}
			s->size = i;
			freeaddrinfo(result);
		}
	}
	if (found) {
		return found;
	}
	return 0;
}

static const response *init(void)
{
	local_ip = getprotoenv("LOCALIP");
	remote_ip = getprotoenv("REMOTEIP");
	local_host = getprotoenv("LOCALHOST");
	remote_host = getprotoenv("REMOTEHOST");
	relayclient = session_getenv("RELAYCLIENT");
	rejectbadhelo = session_getenv("HELO_CHECK_REJECT");
	deferbadhelo = session_getenv("HELO_CHECK_DEFER");
	skiphelocheck = session_getenv("HELO_CHECK_SKIP");
	checkremotehost = session_getenv("HELO_CHECK_REMOTEHOST");
	if (session_getenv("SENDER_HELO_HOSTNAME") != 0) {
		unsetenv("SENDER_HELO_HOSTNAME");
	}
	thispid = (unsigned long)getpid();
	return 0;
}

static const response *helo(str *hostname, str *capabilities)
{
	if (session_getenv("SENDER_HELO_HOSTNAME") == 0) {
		char t[strlen(hostname->s) + 1];
		str_tolower(hostname->s, t);
		session_setenv("SENDER_HELO_HOSTNAME", t, 0);
	}
	return 0;
	(void)capabilities;
}

int check_senderhelo_char_start(const char *s, int i, char c)
{
	char d = '\'';
	if (c == '\'') {
		d = '"';
	}
	/* first character */
	if (*s == c) {
		if (i > 0) {
			char t[strlen(s)+100];
			sprintf(t, "Sorry, your HELO name starts with "
				   "disallowed character %c%c%c", d, c, d);
			resp.message = t;
			return 2;
		} else {
			eprintf_logname();
			eprintf("%c%c%c found in first character of helo "
				"hostname %s\n", d, c, d, s);
			str_cats(&headermsg, "FAILED\n");
			str_cats(&headermsg, "X-HELO-Check-Summary: ");
			str_catc(&headermsg, d);
			str_catc(&headermsg, c);
			str_catc(&headermsg, d);
			str_cats(&headermsg, " found in first character of ");
			str_cats(&headermsg, "helo name ");
			str_cats(&headermsg, s);
			str_cats(&headermsg, "\n");
			return 1;
		}
	}
	return 0;
}

int check_senderhelo_char_end(const char *s, int i, char c)
{
	char d = '\'';
	if (c == '\'') {
		d = '"';
	}
	int len = strlen(s);
	/* last character */
	if (s[len-1] == c) {
		if (i > 0) {
			char t[strlen(s)+100];
			sprintf(t, "Sorry, your HELO name ends with "
				   "disallowed character %c%c%c", d, c, d);
			resp.message = t;
			return 2;
		} else {
			eprintf_logname();
			eprintf("%c%c%c found in last character of helo "
				"hostname %s\n", d, c, d, s);
			str_cats(&headermsg, "FAILED\n");
			str_cats(&headermsg, "X-HELO-Check-Summary: ");
			str_catc(&headermsg, d);
			str_catc(&headermsg, c);
			str_catc(&headermsg, d);
			str_cats(&headermsg, " found in last character of ");
			str_cats(&headermsg, "helo name ");
			str_cats(&headermsg, s);
			str_cats(&headermsg, "\n");
			return 1;
		}
	}
	return 0;
}

int check_senderhelo_char(const char *s, int i, char c)
{
	if (strchr(s, c)) {
		char d = '\'';
		if (c == '\'') {
			d = '"';
		}
		if (i > 0) {
			char t[strlen(s)+100];
			sprintf(t, "Sorry, your HELO name has the disallowed "
				   "character %c%c%c", d, c, d);
			resp.message = t;
			return 2;
		} else {
			eprintf_logname();
			eprintf("%c%c%c character found in helo hostname %s\n",
				d, c, d, s);
			str_cats(&headermsg, "FAILED\n");
			str_cats(&headermsg, "X-HELO-Check-Summary: ");
			str_catc(&headermsg, d);
			str_catc(&headermsg, c);
			str_catc(&headermsg, d);
			str_cats(&headermsg, " character found in helo name ");
			str_cats(&headermsg, s);
			str_cats(&headermsg, "\n");
			return 1;
		}
	}
	return 0;
}

static const response* sender(str* address, str* params)
{
	if (skiphelocheck) return 0;
	if (!local_ip) {
		local_ip = getprotoenv("LOCALIP");
	}
	if (!local_ip) {
		eprintf_logname();
		eprintf("LOCALIP not found thus this plugin skipped\n");
		return 0;
	}
	local_host = getprotoenv("LOCALHOST");
	if (!remote_ip) {
		remote_ip = getprotoenv("REMOTEIP");
	}
	if (!remote_ip) {
		eprintf_logname();
		eprintf("REMOTEIP not found thus this plugin skipped\n");
		return 0;
	}
	if (!remote_host) {
		remote_host = getprotoenv("REMOTEHOST");
	}
	if (!remote_host) {
		eprintf_logname();
		eprintf("REMOTEHOST is empty thus trying to do PTR DNS lookup "
			"from REMOTEIP %s\n", remote_ip);
		const char *tmp = ip_ptr_lookup((char *)remote_ip);
		if (strcmp(tmp, remote_ip) != 0) {
			remote_host = tmp;
			eprintf_logname();
			eprintf("PTR DNS loopkup of REMOTEIP=%s success and "
				"set REMOTEHOST=%s\n", remote_ip, remote_host);
		}
	}
	str_init(&headermsg);
	str_copys(&headermsg, "X-HELO-Check-Result: ");
	if (session_getnum("authenticated", 0)) {
		eprintf_logname();
		eprintf("authenticated user thus skipped\n");
		str_cats(&headermsg, "SKIPPED: authenticated user\n");
		return 0;
	}
	if (relayclient) {
		eprintf_logname();
		eprintf("$RELAYCLIENT is set thus skipped\n");
		str_cats(&headermsg, "SKIPPED: $RELAYCLIENT is set\n");
		return 0;
	}
	if (strcmp(local_ip, remote_ip) == 0) {
		eprintf_logname();
		eprintf("LOCALIP=REMOTEIP=%s\n", remote_ip);
		str_cats(&headermsg, "SKIPPED: LOCALIP is the same as "
				     "REMOTEIP ");
		str_cats(&headermsg, remote_ip);
		str_cats(&headermsg, "\n");
		return 0;
	}
	int todeferreject = 0;
	if (rejectbadhelo || deferbadhelo) {
		if (rejectbadhelo) {
			todeferreject = 554;
			resp.number = todeferreject;
		} else {
			todeferreject = 451;
			resp.number = todeferreject;
		}
	}
	if (!remote_host) {
		if (todeferreject > 0) {
			str_init(&respmsg);
			str_cats(&respmsg, "Sorry, your hostname is empty and "
					   "your ip ");
			str_cats(&respmsg, remote_ip);
			str_cats(&respmsg, " doesn't have any PTR record");
			resp.message = respmsg.s;
			return &resp;
		} else {
			eprintf_logname();
			eprintf("REMOTEHOST is empty and PTR DNS lookup "
				"for REMOTEIP %s failed\n", remote_ip);
			str_cats(&headermsg, "FAILED\n");
			str_cats(&headermsg, "X-HELO-Check-Summary: "
					     "REMOTEHOST is empty and PTR DNS "
					     " lookup for IP ");
			str_cats(&headermsg, remote_ip);
			str_cats(&headermsg, " failed\n");
			return 0;
		}
	}
	const char *senderhelo = session_getenv("SENDER_HELO_HOSTNAME");
	/* If helo_hostname is empty */
	if (senderhelo == 0) {
		if (todeferreject > 0) {
			resp.message = "Sorry, your HELO name is required";
			return &resp;
		} else {
			eprintf_logname();
			eprintf("helo hostname is empty\n");
			str_cats(&headermsg, "FAILED\n");
			str_cats(&headermsg, "X-HELO-Check-Summary: HELO name"
					     " is empty!\n");
			return 0;
		}
	}
	if (senderhelo) {
		if (checkremotehost) {
			eprintf_logname();
			eprintf("$HELO_CHECK_REMOTEHOST is set\n");
			if (strcmp(senderhelo, remote_host) != 0) {
				if (todeferreject > 0) {
					str_init(&respmsg);
					str_cats(&respmsg, "Sorry, your HELO "
							   "name ");
					str_cats(&respmsg, senderhelo);
					str_cats(&respmsg, " and remotehost ");
					str_cats(&respmsg, remote_host);
					str_cats(&respmsg, " doesn't match");
					resp.message = respmsg.s;
					return &resp;
				} else {
					eprintf_logname();
					eprintf("helohostname=%s remotehost=%s"
						" don't match\n",
						senderhelo, remote_host);
					str_cats(&headermsg, "FAILED\n");
					str_cats(&headermsg, "X-HELO-Check-"
							     "Summary: helo "
							     "hostname ");
					str_cats(&headermsg, senderhelo);
					str_cats(&headermsg, " remotehost ");
					str_cats(&headermsg, remote_host);
					str_cats(&headermsg, " doesn't "
							     "match\n");
					return 0;
				}
			} else {
				eprintf_logname();
				eprintf("helohostname=%s remotehost=%s "
					"match\n", senderhelo, remote_host);
			}
		} else {
			eprintf_logname();
			eprintf("$HELO_CHECK_REMOTEHOST is not set\n");
		}
		int len = strlen(senderhelo);
		/* 0 continue, 1 to return 0 and 2 to return &resp; */
		int checkret = 0;
		/* Don't accept start with a dot */
		checkret = check_senderhelo_char_start(senderhelo,
						       todeferreject, '.');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		/* Don't accept start with a [ */
		checkret = check_senderhelo_char_start(senderhelo,
						       todeferreject, '[');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		/* Don't accept start with a dash */
		checkret = check_senderhelo_char_start(senderhelo,
						       todeferreject, '-');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		if (len > 1) {
			/* Don't accept end with a dot */
			checkret = check_senderhelo_char_end(senderhelo,
							     todeferreject,
							     '.');
			if (checkret == 2) return &resp;
			if (checkret == 1) return 0;
			/* Don't accept end with a ] */
			checkret = check_senderhelo_char_end(senderhelo,
							     todeferreject,
							     ']');
			if (checkret == 2) return &resp;
			if (checkret == 1) return 0;
			/* Don't accept end with a dash */
			checkret = check_senderhelo_char_end(senderhelo,
							     todeferreject,
							     '-');
			if (checkret == 2) return &resp;
			if (checkret == 1) return 0;
		}
		if (strchr(senderhelo, '.') == NULL) {
			if (todeferreject > 0) {
				resp.message = "Sorry, your HELO name doesn't "
					       "contains the character '.'";
				return &resp;
			} else {
				eprintf_logname();
				eprintf("'.' character not found in helo "
					"hostname %s\n", senderhelo);
				str_cats(&headermsg, "FAILED\n");
				str_cats(&headermsg, "X-HELO-Check-Summary: "
						     "'.' character not found "
						     "in helo name ");
				str_cats(&headermsg, senderhelo);
				str_cats(&headermsg, "\n");
				return 0;
			}
		}
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '!');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '@');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '#');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '$');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '%');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '^');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '&');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '*');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '(');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 ')');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '[');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 ']');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '{');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '}');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 ':');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 ';');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '"');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '\'');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '<');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '>');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '?');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '/');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '|');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '\\');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '+');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '=');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '_');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 ' ');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 '\t');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		checkret = check_senderhelo_char(senderhelo, todeferreject,
						 ',');
		if (checkret == 2) return &resp;
		if (checkret == 1) return 0;
		int okipv4 = is_valid_ipv4(senderhelo);
		if (okipv4 > 0) {
			if (todeferreject > 0) {
				resp.message = "Sorry, your HELO name is a "
					       "valid IPv4 address which is "
					       "disallowed";
				return &resp;
			} else {
				eprintf_logname();
				eprintf("valid IPv4 address found in helo "
					"hostname %s\n", senderhelo);
				str_cats(&headermsg, "FAILED\n");
				str_cats(&headermsg, "X-HELO-Check-Summary: "
						     "Valid IPv4 address found"
						     " in helo name ");
				str_cats(&headermsg, senderhelo);
				str_cats(&headermsg, "\n");
				return 0;
			}
		}
	}
	/* 
	 * TODO: We need to check localhost, localhost.localdomain etc.
	 * Reference: http://www.linuxmagic.com/best_practices/valid_helo_domain.html
	 * Maybe implement reading files from QMAILHOME/control folders for
	 * rcpthosts, me, locals, etc. and if match will reject/defer or tag?
	 */
	struct HostDNSIP hh;
	hh.size = 0;
	hh.ptrcount = 0;
	hh.ptrcheck = 0;
	hh.hostname = NULL;
	hh.ipmemused = 0;
	hh.ptrmemused = 0;
	eprintf_logname();
	eprintf("Start checking heloname=%s\n", senderhelo);
	int ret = check_hostname(senderhelo, &hh);
	if (ret > 0) {
		if (hh.size > 0) {
			int x = 0;
			for (x = 0; x < hh.size && hh.ip[x] != NULL; x++) {
				eprintf_logname();
				eprintf("ip[%d]=%s", x, hh.ip[x]);
				eprintf(" ptr[%d]=%s", x, hh.ptr[x]);
				/* If it is different then it is a ptr */
				if (strcmp(hh.ip[x], hh.ptr[x]) != 0) {
					hh.ptrcount++;
					eprintf(" ptr=YES");
				} else {
					eprintf(" ptr=NO");
				}
				/* If it is the same then it is valid ptr */
				if (strcmp(hh.hostname, hh.ptr[x]) == 0) {
					hh.ptrcheck++;
					eprintf(" check=PASSED\n");
				} else {
					eprintf(" check=FAILED\n");
				}
			}
		}
		const char *helo_ptrcheck = session_getenv("HELO_CHECK_PTR");
		if (hh.size > 0) {
			if (hh.ptrcount == 0) {
				if (todeferreject > 0) {
					resp.message = "Sorry, your HELO name "
						       "doesn't has a valid "
						       "PTR record";
					(void)free_HostDNSIP(&hh);
					return &resp;
				} else {
					str_cats(&headermsg, "FAILED\n");
					str_cats(&headermsg, "X-HELO-Check-"
							     "Summary: PTR "
							     "record not "
							     "found for helo"
							     " name ");
					str_cats(&headermsg, senderhelo);
					str_cats(&headermsg, "\n");
					(void)free_HostDNSIP(&hh);
					return 0;
				}
			}
			if (hh.ptrcheck == 0 && helo_ptrcheck) {
				if (todeferreject > 0) {
					resp.message = "Sorry, your HELO name "
						       "doesn't has a valid "
						       "PTR record that "
						       "resolves back to the "
						       "same name";
					(void)free_HostDNSIP(&hh);
					return &resp;
				} else {
					str_cats(&headermsg, "FAILED\n");
					str_cats(&headermsg, "X-HELO-Check-"
							     "Summary: PTR "
							     "record doesn't "
							     "resolves back "
							     "to helo name ");
					str_cats(&headermsg, senderhelo);
					str_cats(&headermsg, "\n");
					(void)free_HostDNSIP(&hh);
					return 0;
				}
			}
		}
		if (hh.ptrcheck > 0) {
			str_cats(&headermsg, "PASSED\n");
		} else {
			str_cats(&headermsg, "FAILED\n");
		}
		char tt[512];
		if (hh.size > 1) {
			sprintf(tt, "X-HELO-Check-Summary: There are "
				    "%d A records for %s",
				    hh.size, hh.hostname);
			str_cats(&headermsg, tt);
		} else {
			sprintf(tt, "X-HELO-Check-Summary: There is "
				    "%d A record for %s",
				    hh.size, hh.hostname);
			str_cats(&headermsg, tt);
		}
		if (hh.ptrcount > 1) {
			sprintf(tt, ", %d PTR records and ",
				    hh.ptrcount);
			str_cats(&headermsg, tt);
		} else {
			sprintf(tt, ", %d PTR record and ",
				    hh.ptrcount);
			str_cats(&headermsg, tt);
		}
		if (hh.ptrcheck > 1) {
			sprintf(tt, "%d PTR records resolve to the "
				    "same hostname.\n", hh.ptrcheck);
			str_cats(&headermsg, tt);
		} else {
			sprintf(tt, "%d PTR record resolves to the "
				    "same hostname.\n", hh.ptrcheck);
			str_cats(&headermsg, tt);
		}
		(void)free_HostDNSIP(&hh);
		return 0;
	} else {
		if (outofmemory > 0) {
			return &resp_oom;
		}
		if (todeferreject > 0) {
			resp.message = "Sorry, your HELO name is invalid";
			if (errormsg.len > 0) {
				resp.message = errormsg.s;
			}
			return &resp;
		} else {
			eprintf_logname();
			str_cats(&headermsg, "FAILED\n");
			if (errormsg.len > 0) {
				eprintf("(%s:%d) helo hostname %s DNS lookup "
					"error: %s\n", __FILE__, __LINE__,
					senderhelo, errormsg.s);
				str_cats(&headermsg, "X-HELO-Check-Summary: "
						     "DNS lookup error for "
						     "helo name ");
				str_cats(&headermsg, senderhelo);
				str_cats(&headermsg, ": ");
				str_cats(&headermsg, errormsg.s);
				str_cats(&headermsg, "\n");
			} else {
				eprintf("(%s:%d) helo hostname %s DNS lookup "
					"error\n",
					__FILE__, __LINE__, senderhelo);
				str_cats(&headermsg, "X-HELO-Check-Summary: "
						     "DNS lookup error for "
						     "helo name ");
				str_cats(&headermsg, senderhelo);
				str_cats(&headermsg, "\n");
			}
			return 0;
		}
	}
	return 0;
	(void)address;
	(void)params;
}

static const response* data_start(int fd)
{
	if (skiphelocheck) return 0;
	if (headermsg.len > 0) {
		return backend_data_block(headermsg.s, headermsg.len);
	}
	return 0;
	(void)fd;
}

struct plugin plugin = {
	.version = PLUGIN_VERSION,
	.flags = 0,
	.init = init,
	.helo = helo,
	.sender = sender,
	.data_start = data_start,
};